package com.sd.business.service.inventory.impl;

import cn.hutool.core.date.DateUtil;
import cn.hutool.core.util.ObjectUtil;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.sd.business.entity.bom.po.Bom;
import com.sd.business.entity.bom.po.BomSpec;
import com.sd.business.entity.department.constant.DepartmentConstant;
import com.sd.business.entity.department.po.Department;
import com.sd.business.entity.inventory.dto.InventorySelectDto;
import com.sd.business.entity.inventory.dto.QuantityByDepartmentDto;
import com.sd.business.entity.inventory.dto.QuantityByWarehouseDto;
import com.sd.business.entity.inventory.dto.QuantityDto;
import com.sd.business.entity.inventory.po.Inventory;
import com.sd.business.entity.inventory.vo.InventoryVo;
import com.sd.business.entity.inventory.vo.QuantityByDepartmentVo;
import com.sd.business.entity.inventory.vo.QuantityByWarehouseVo;
import com.sd.business.entity.order.po.OrderSku;
import com.sd.business.entity.warehouse.constant.WarehouseConstant;
import com.sd.business.entity.warehouse.po.Warehouse;
import com.sd.business.mapper.inventory.InventoryMapper;
import com.sd.business.service.inventory.InventoryService;
import com.sd.framework.util.excel.util.ExcelUtil;
import com.sd.framework.util.sql.Sql;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import javax.servlet.http.HttpServletResponse;
import java.math.BigDecimal;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;


/**
 * <p>
 * 库存 服务实现类
 * </p>
 *
 * @author
 * @since 2023-12-11
 */
@Service
public class InventoryServiceImpl extends ServiceImpl<InventoryMapper, Inventory> implements InventoryService {

    @Autowired
    private HttpServletResponse response;

    @Override
    public Page<InventoryVo> getPage(InventorySelectDto dto) {

        return Sql.create(InventoryVo.class)
                .selectAll(Inventory.class)
                .selectAs(Warehouse::getName, InventoryVo::getWarehouseName)
                .selectAs(BomSpec::getCode, InventoryVo::getBomSpecCode)
                .selectAs(BomSpec::getName, InventoryVo::getBomSpecName)
                .selectAs(Department::getName, InventoryVo::getDepartmentName)
                .multiply(Inventory::getQuantity, Inventory::getBalanceUnitPrice, InventoryVo::getTotalPrice)
                .add(Inventory::getQuantity, Inventory::getLockQuantity, InventoryVo::getTotalQuantity)

                .from(Inventory.class)
                .leftJoin(BomSpec.class, BomSpec::getId, Inventory::getBomSpecId)
                .leftJoin(Bom.class, Bom::getId, BomSpec::getBomId)
                .leftJoin(Warehouse.class, Warehouse::getId, Inventory::getWarehouseId)
                .leftJoin(Department.class, Department::getId, Inventory::getDepartmentId)

                .eq(Inventory::getDepartmentId, dto.getDepartmentId())
                .eq(Inventory::getWarehouseId, dto.getWarehouseId())
                .in(Inventory::getWarehouseId, dto.getWarehouseIdList())
                .like(BomSpec::getName, dto.getBomSpecName())
                .like(BomSpec::getCode, dto.getBomSpecCode())
                .like(BomSpec::getColour, dto.getBomSpecColour())
                .like(BomSpec::getLength, dto.getBomSpecLength())
                .like(BomSpec::getWidth, dto.getBomSpecWidth())
                .eq(Bom::getBomClassifyId, dto.getBomClassifyId())
                .eq(Bom::getSpecies, dto.getBomSpecies())
                .eq(Bom::getChromatophore, dto.getBomChromatophore())
                .eq(Bom::getEmbossingProcess, dto.getBomEmbossingProcess())
                .eq(Bom::getFrontGrain, dto.getBomFrontGrain())
                .eq(Bom::getReverseGrain, dto.getBomReverseGrain())
                .orderByDesc(Inventory::getId)
                .page(dto);
    }

    @Override
    public void exportExcel(InventorySelectDto dto) {
        dto.setSearchAll(true);
        Page<InventoryVo> page = getPage(dto);
        List<InventoryVo> list = page.getRecords();
        ExcelUtil.export(response, DateUtil.format(new Date(), "yyyy-MM-dd库存"), "库存", list, InventoryVo.class);
    }

    @Override
    public Map<Long, BigDecimal> getQuantity(QuantityDto dto) {
        return mapKV(Inventory::getBomSpecId, Inventory::getQuantity, q -> q
                .eq(Inventory::getWarehouseId, dto.getWarehouseId())
                .eq(Inventory::getDepartmentId, dto.getDepartmentId())
                .in(Inventory::getBomSpecId, dto.getBomSpecIdList()));
    }

    @Override
    public List<QuantityByWarehouseVo> getQuantityByWarehouse(QuantityByWarehouseDto dto) {

        return Sql.create(QuantityByWarehouseVo.class)
                .sum(Inventory::getQuantity, QuantityByWarehouseVo::getInventoryQuantity)
                .selectAs(Warehouse::getId, QuantityByWarehouseVo::getWarehouseId)
                .selectAs(Warehouse::getName, QuantityByWarehouseVo::getWarehouseName)

                .from(Inventory.class)
                .innerJoin(Warehouse.class, Warehouse::getId, Inventory::getId)

                .eq(Inventory::getDepartmentId, dto.getDepartmentId())
                .groupBy(Warehouse::getId)
                .orderByAsc(Warehouse::getId)
                .list();
    }

    @Override
    public List<QuantityByDepartmentVo> getQuantityByDepartment(QuantityByDepartmentDto dto) {

        return Sql.create(QuantityByDepartmentVo.class)
                .sum(Inventory::getQuantity, QuantityByDepartmentVo::getInventoryQuantity)
                .selectAs(Department::getId, QuantityByDepartmentVo::getDepartmentId)
                .selectAs(Department::getName, QuantityByDepartmentVo::getDepartmentName)

                .from(Inventory.class)
                .innerJoin(Department.class, Department::getId, Inventory::getDepartmentId)

                .eq(Inventory::getWarehouseId, dto.getWarehouseId())
                .groupBy(Department::getId)
                .orderByAsc(Department::getId)
                .list();
    }

    @Override
    public boolean lockStorage(Map<Long, BigDecimal> map, Long warehouseId) {
        if (ObjectUtil.isEmpty(map)) {
            return true;
        }
        synchronized (this) {
            List<Inventory> list = list(q -> q
                    .in(Inventory::getBomSpecId, map.keySet())
                    .eq(Inventory::getWarehouseId, warehouseId)
                    .eq(Inventory::getDepartmentId, DepartmentConstant.SD_SPORTS)
            );
            if (map.size() != list.size()) {
                return false;
            }
            for (Inventory inventory : list) {
                BigDecimal lockQuantity = map.get(inventory.getBomSpecId());
                if (inventory.getQuantity().compareTo(lockQuantity) >= 0) {
                    inventory.setQuantity(inventory.getQuantity().subtract(lockQuantity));
                    inventory.setLockQuantity(inventory.getLockQuantity().add(lockQuantity));
                } else {
                    return false;
                }
            }
            this.updateBatchById(list);
        }
        return true;
    }

    @Override
    public boolean lockStorage(List<OrderSku> orderSkuList) {
        Map<Long, BigDecimal> map = orderSkuList.stream()
                .collect(Collectors.toMap(OrderSku::getBomSpecId, OrderSku::getQuantity, BigDecimal::add));
        return lockStorage(map, WarehouseConstant.SEMI_FINISHED_PRODUCT);
    }

    @Override
    public void unlockStorage(Map<Long, BigDecimal> map, Long warehouseId) {
        if (ObjectUtil.isEmpty(map)) {
            return;
        }
        synchronized (this) {
            List<Inventory> list = list(q -> q
                    .in(Inventory::getBomSpecId, map.keySet())
                    .eq(Inventory::getWarehouseId, warehouseId)
                    .eq(Inventory::getDepartmentId, DepartmentConstant.SD_SPORTS)
            );
            for (Inventory inventory : list) {
                BigDecimal lockQuantity = map.get(inventory.getBomSpecId());
                inventory.setQuantity(inventory.getQuantity().add(lockQuantity));
                inventory.setLockQuantity(inventory.getLockQuantity().subtract(lockQuantity));
            }
            updateBatchById(list);
        }
    }

    @Override
    public void unlockStorage(List<OrderSku> orderSkuList) {
        if (orderSkuList.isEmpty()) {
            return;
        }
        Map<Long, BigDecimal> map = orderSkuList.stream()
                .collect(Collectors.toMap(OrderSku::getBomSpecId, OrderSku::getQuantity, BigDecimal::add));
        unlockStorage(map, WarehouseConstant.SEMI_FINISHED_PRODUCT);
    }


}
